import logging
import time
from typing import List, Union

import numpy as np
from BoltzTraP2.bandlib import calc_Onsager_coefficients

from amset.constants import bohr_to_cm, defaults, e_si
from amset.core.data import AmsetData
from amset.interpolation.boltztrap import fermiintegrals
from amset.log import log_time_taken
from amset.util import get_progress_bar

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

logger = logging.getLogger(__name__)

_e_str = "Electronic structure must contain dopings, temperatures, and scattering rates"


def solve_boltzman_transport_equation(
    amset_data: AmsetData,
    calculate_mobility: bool = defaults["calculate_mobility"],
    separate_mobility: bool = defaults["separate_mobility"],
    progress_bar: bool = defaults["print_log"],
):
    has_doping = amset_data.doping is not None
    has_temps = amset_data.temperatures is not None
    has_rates = amset_data.scattering_rates is not None
    if not (has_doping and has_temps and has_rates):
        raise ValueError(_e_str)

    logger.info(
        "Calculating conductivity, Seebeck, and electronic thermal conductivity"
    )
    t0 = time.perf_counter()
    sigma, seebeck, kappa = _calculate_transport_properties(
        amset_data, progress_bar=progress_bar
    )
    log_time_taken(t0)

    if not calculate_mobility:
        return sigma, seebeck, kappa, None

    if amset_data.is_metal:
        logger.info("System is metallic, refusing to calculate carrier mobility")
        return sigma, seebeck, kappa, None

    n_scats = len(amset_data.scattering_labels)

    logger.info("Calculating overall mobility")
    t0 = time.perf_counter()
    overall = _calculate_mobility(
        amset_data, np.arange(n_scats), pbar_label="mobility" if progress_bar else None
    )
    mobility = {"overall": overall}
    log_time_taken(t0)

    if separate_mobility:
        logger.info("Calculating individual scattering rate mobilities")
        t0 = time.perf_counter()
        for rate_idx, name in enumerate(amset_data.scattering_labels):
            mobility[name] = _calculate_mobility(
                amset_data, rate_idx, pbar_label=name if progress_bar else None
            )
        log_time_taken(t0)

    return sigma, seebeck, kappa, mobility


def _calculate_mobility(
    amset_data: AmsetData,
    rate_idx: Union[int, List[int], np.ndarray],
    pbar_label: str = "mobility",
):
    if isinstance(rate_idx, int):
        rate_idx = [rate_idx]

    volume = amset_data.structure.volume
    mobility = np.zeros(amset_data.fermi_levels.shape + (3, 3))

    epsilon, dos = amset_data.tetrahedral_band_structure.get_density_of_states(
        amset_data.dos.energies, sum_spins=True, use_cached_weights=True
    )

    if pbar_label is not None:
        pbar = get_progress_bar(
            iterable=list(np.ndindex(amset_data.fermi_levels.shape)), desc=pbar_label
        )
    else:
        pbar = list(np.ndindex(amset_data.fermi_levels.shape))

    for n, t in pbar:
        br = {s: np.arange(len(amset_data.energies[s])) for s in amset_data.spins}
        cb_idx = {s: amset_data.vb_idx[s] + 1 for s in amset_data.spins}

        if amset_data.doping[n] < 0:
            band_idx = {s: br[s][cb_idx[s] :] for s in amset_data.spins}
        else:
            band_idx = {s: br[s][: cb_idx[s]] for s in amset_data.spins}

        lifetimes = {
            s: 1 / np.sum(amset_data.scattering_rates[s][rate_idx, n, t], axis=0)
            for s in amset_data.spins
        }

        # Nones are required as BoltzTraP2 expects the Fermi and temp as arrays
        fermi = amset_data.fermi_levels[n, t][None]
        temp = amset_data.temperatures[t][None]

        # obtain the Fermi integrals for the temperature and doping
        vvdos = get_transport_dos(
            amset_data.tetrahedral_band_structure,
            amset_data.velocities_product,
            lifetimes,
            amset_data.dos.energies,
            band_idx=band_idx,
        )

        c, l0, l1, l2, lm11 = fermiintegrals(
            epsilon, dos, vvdos, mur=fermi, Tr=temp, dosweight=amset_data.dos.dos_weight
        )

        # Compute the Onsager coefficients from Fermi integrals
        sigma, _, _, _ = calc_Onsager_coefficients(l0, l1, l2, fermi, temp, volume)

        if amset_data.doping[n] < 0:
            carrier_conc = amset_data.electron_conc[n, t]
        else:
            carrier_conc = amset_data.hole_conc[n, t]

        # don't use c as we don't use the correct DOS each time
        # c = -c[0, ...] / (volume / (Meter / 100.)**3)

        # convert mobility to cm^2/V.s
        uc = 0.01 / (e_si * carrier_conc * (1 / bohr_to_cm) ** 3)
        mobility[n, t] = sigma[0, ...] * uc

    return mobility


def _calculate_transport_properties(
    amset_data, progress_bar: bool = defaults["print_log"]
):
    n_t_size = (len(amset_data.doping), len(amset_data.temperatures))

    sigma = np.zeros(n_t_size + (3, 3))
    seebeck = np.zeros(n_t_size + (3, 3))
    kappa = np.zeros(n_t_size + (3, 3))
    volume = amset_data.structure.volume

    epsilon, dos = amset_data.tetrahedral_band_structure.get_density_of_states(
        amset_data.dos.energies, sum_spins=True, use_cached_weights=True
    )

    iterable = list(np.ndindex(n_t_size))
    if progress_bar:
        pbar = get_progress_bar(iterable=iterable, desc="transport")
    else:
        pbar = iterable

    # solve sigma, seebeck, kappa and hall using information from all bands
    for n, t in pbar:
        lifetimes = {
            s: 1 / np.sum(amset_data.scattering_rates[s][:, n, t], axis=0)
            for s in amset_data.spins
        }

        # Nones are required as BoltzTraP2 expects the Fermi and temp as arrays
        fermi = amset_data.fermi_levels[n, t][None]
        temp = amset_data.temperatures[t][None]

        # obtain the Fermi integrals
        vvdos = get_transport_dos(
            amset_data.tetrahedral_band_structure,
            amset_data.velocities_product,
            lifetimes,
            amset_data.dos.energies,
        )

        _, l0, l1, l2, lm11 = fermiintegrals(
            epsilon, dos, vvdos, mur=fermi, Tr=temp, dosweight=amset_data.dos.dos_weight
        )

        # Compute the Onsager coefficients from Fermi integrals
        # Don't store the Hall coefficient as we don't have the curvature
        # information.
        sigma[n, t], seebeck[n, t], kappa[n, t], _ = calc_Onsager_coefficients(
            l0, l1, l2, fermi, temp, volume
        )

    # convert seebeck to µV/K
    seebeck *= 1e6

    return sigma, seebeck, kappa


def get_transport_dos(
    tetrahedron_band_structure, vvband, lifetimes, energies, band_idx=None
):
    """Compute the transport DOS

    The transport DOS is weighted by the outer product of the group velocity
    with itself, and by the relaxation time.

    Args:
        vvband: (nbands, 3, 3, nkpoints) array with the outer product of each
            group velocity with itself.

    Returns:
        The transport dos with the same (3, 3, npts).
    """
    # vvband is nb, 3, 3, nk it should be nb, nk, 3, 3
    vvband = {s: v.transpose(0, 3, 1, 2) for s, v in vvband.items()}
    weights = {s: vvband[s] * lifetimes[s][:, :, None, None] for s in lifetimes}

    _, vvdos = tetrahedron_band_structure.get_density_of_states(
        energies,
        integrand=weights,
        sum_spins=True,
        band_idx=band_idx,
        use_cached_weights=True,
    )

    # vvdos is npts, 3, 3 it should be 3, 3, npts
    vvdos = vvdos.transpose(1, 2, 0)

    return vvdos



from copy import deepcopy
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np

from amset.plot import styled_plot
from amset.plot.base import BaseMeshPlotter, amset_base_style

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

_legend_kwargs = {"loc": "upper left", "bbox_to_anchor": (1, 1), "frameon": False}


class RatesPlotter(BaseMeshPlotter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.plot_energies = np.vstack([self.energies[s] for s in self.spins])
        self.plot_rates = np.concatenate(
            [self.scattering_rates[s] for s in self.spins], axis=3
        )

    @styled_plot(amset_base_style)
    def get_plot(
        self,
        plot_fd_tols: bool = False,
        plot_total_rate: bool = False,
        ymin: float = None,
        ymax: float = None,
        xmin: float = None,
        xmax: float = None,
        normalize_energy: bool = True,
        separate_rates: bool = True,
        doping_idx=0,
        temperature_idx=0,
        legend=True,
        style=None,
        no_base_style=False,
        fonts=None,
    ):
        if normalize_energy and self.is_metal:
            norm_e = self.fermi_levels[0][0]
        elif normalize_energy:
            cb_idx = {s: v + 1 for s, v in self.vb_idx.items()}
            norm_e = np.max([self.energies[s][: cb_idx[s]] for s in self.spins])
        else:
            norm_e = 0

        if doping_idx is None and len(self.doping) == 1:
            doping_idx = 0

        if temperature_idx is None and len(self.temperatures) == 1:
            temperature_idx = 0

        if doping_idx is None and temperature_idx is None:
            ny = len(self.doping)
            nx = len(self.temperatures)
            p_idx = [[(x, y) for y in range(nx)] for x in range(ny)]
        elif doping_idx is None:
            nx = len(self.doping)
            ny = 1
            p_idx = [[(x, temperature_idx) for x in range(nx)]]
        elif temperature_idx is None:
            nx = len(self.temperatures)
            ny = 1
            p_idx = [[(doping_idx, x) for x in range(nx)]]
        else:
            nx = 1
            ny = 1
            p_idx = [[(doping_idx, temperature_idx)]]

        base_size = 3.2

        def get_size(nplots):
            width = base_size * nplots
            width += (
                base_size * 0.25 * (nplots - 1)
            )  # account for spacing between plots
            return width

        if ny == 1:
            figsize = (get_size(nx), base_size)
            fig, axes = plt.subplots(1, nx, figsize=figsize)
        else:
            figsize = (get_size(nx), get_size(ny))
            fig, axes = plt.subplots(ny, nx, figsize=figsize)

        for y_idx, p_x_idx in enumerate(p_idx):
            for x_idx, (n, t) in enumerate(p_x_idx):
                if nx == 1 and ny == 1:
                    ax = axes
                elif ny == 1:
                    ax = axes[x_idx]
                else:
                    ax = axes[y_idx, x_idx]

                if x_idx == nx - 1 and y_idx == ny - 1 and legend:
                    show_legend = True
                else:
                    show_legend = False

                str_doping = _latex_float(self.doping[n])

                title = r"n = ${}$ cm$^{{-3}}$    " "T = {} K".format(
                    str_doping, self.temperatures[t]
                )

                self.plot_rates_to_axis(
                    ax,
                    n,
                    t,
                    separate_rates=separate_rates,
                    plot_total_rate=plot_total_rate,
                    plot_fd_tols=plot_fd_tols,
                    ymin=ymin,
                    ymax=ymax,
                    xmin=xmin,
                    xmax=xmax,
                    show_legend=show_legend,
                    legend_kwargs=_legend_kwargs,
                    normalize_energy=norm_e,
                )
                ax.set(title=title)

        return plt

    def plot_rates_to_axis(
        self,
        ax,
        doping_idx,
        temperature_idx,
        separate_rates=True,
        plot_total_rate: bool = False,
        plot_fd_tols: bool = True,
        show_legend: bool = True,
        legend_kwargs=None,
        ymin=None,
        ymax=None,
        xmin=None,
        xmax=None,
        normalize_energy=0,
        plot_lifetimes=False,
        scaling_factor=1,
    ):
        rates = self.plot_rates[:, doping_idx, temperature_idx]
        if separate_rates:
            labels = self.scattering_labels.tolist()
        else:
            rates = np.sum(rates, axis=0)[None, ...]
            labels = ["overall"]

        if legend_kwargs is None:
            legend_kwargs = {}

        labels = deepcopy(labels)
        if plot_total_rate:
            # add total rates column
            rates = np.concatenate((rates, np.sum(rates, axis=0)[None, ...]))
            labels += ["total"]

        min_fd = self.fd_cutoffs[0]
        max_fd = self.fd_cutoffs[1]
        if not xmin:
            min_e = min_fd
        else:
            min_e = xmin

        if not xmax:
            max_e = max_fd
        else:
            max_e = xmax

        energies = self.plot_energies.ravel()
        energy_mask = (energies > min_e) & (energies < max_e)
        energies = energies[energy_mask]

        plt_rates = {}
        for label, rate in zip(labels, rates):
            rate = rate.ravel()
            if plot_lifetimes:
                rate = 1 / rate
            rate *= scaling_factor
            plt_rates[label] = rate[energy_mask]

        rates_in_cutoffs = np.array(list(plt_rates.values()))

        ymin, ymax = _get_rate_ylims(rates_in_cutoffs, ymin=ymin, ymax=ymax)

        # convert energies to eV and normalise
        norm_energies = energies - normalize_energy
        norm_min_fd = min_fd - normalize_energy
        norm_max_fd = max_fd - normalize_energy
        norm_min_e = min_e - normalize_energy
        norm_max_e = max_e - normalize_energy

        for label, rate in plt_rates.items():
            ax.scatter(norm_energies, rate, label=label)

        if plot_fd_tols:
            ax.plot(
                (norm_min_fd, norm_min_fd), (ymin, ymax), c="gray", ls="--", alpha=0.5
            )
            ax.plot(
                (norm_max_fd, norm_max_fd),
                (ymin, ymax),
                c="gray",
                ls="--",
                alpha=0.5,
                label="FD cutoffs",
            )

        ax.semilogy()
        ax.set_ylim(ymin, ymax)
        ax.set_xlim((norm_min_e, norm_max_e))
        scaling_string = "{:g} ".format(scaling_factor) if scaling_factor != 1 else ""
        if plot_lifetimes:
            ylabel = "Scattering lifetime ({}s)".format(scaling_string)
        else:
            ylabel = "Scattering rate ({}s$^{{-1}}$)".format(scaling_string)

        ax.set(xlabel="Energy (eV)", ylabel=ylabel)

        if show_legend:
            ax.legend(**legend_kwargs)


def _get_rate_ylims(
    rates, ymin: Optional[float] = None, ymax: Optional[float] = None, pad: float = 0.1
):
    rates = rates[np.isfinite(rates)]

    if not ymin:
        min_log = np.log10(np.min(rates))
    else:
        min_log = np.log10(ymin)

    if not ymax:
        max_log = np.log10(np.max(rates))
    else:
        max_log = np.log10(ymax)

    diff_log = max_log - min_log

    if not ymin:
        log_tmp = min_log - diff_log * pad
        ymin = 10 ** log_tmp

    if not ymax:
        log_tmp = max_log + diff_log * pad
        ymax = 10 ** log_tmp

    return ymin, ymax


def _latex_float(f):
    float_str = "{0:.2g}".format(f)
    if "e" in float_str:
        base, exponent = float_str.split("e")
        return r"{0} x 10^{{{1}}}".format(base, int(exponent))
    else:
        return float_str



import logging

import numpy as np

from amset.constants import defaults
from amset.electronic_structure.kpoints import get_mesh_from_kpoint_numbers
from amset.electronic_structure.symmetry import expand_kpoints
from amset.interpolation.periodic import (
    PeriodicLinearInterpolator,
    group_bands_and_kpoints,
)
from amset.util import array_from_buffer, create_shared_array
from amset.wavefunction.common import desymmetrize_coefficients, get_overlap, is_ncl
from amset.wavefunction.io import load_coefficients

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

logger = logging.getLogger(__name__)


class WavefunctionOverlapCalculator(PeriodicLinearInterpolator):
    def __init__(self, nbands, data_shape, interpolators, ncl, gpoints):
        super().__init__(nbands, data_shape, interpolators)
        self.ncl = ncl
        self.gpoints = gpoints

    def to_reference(self):
        interpolator_references = self._interpolators_to_reference()
        gpoints_buffer, self.gpoints = create_shared_array(
            self.gpoints, return_shared_data=True
        )
        return (
            self.nbands,
            self.data_shape,
            interpolator_references,
            self.ncl,
            gpoints_buffer,
        )

    @classmethod
    def from_reference(
        cls, nbands, data_shape, interpolator_references, ncl, gpoints_buffer
    ):
        interpolators = cls._interpolators_from_reference(interpolator_references)
        gpoints = array_from_buffer(gpoints_buffer)
        return cls(nbands, data_shape, interpolators, ncl, gpoints)

    @classmethod
    def from_data(cls, kpoints, data, gpoints=None, gaussian=None):
        if gpoints is None:
            raise ValueError("gpoints required for initialization")
        ncl = is_ncl(data)
        grid_kpoints, mesh_dim, sort_idx = cls._grid_kpoints(kpoints)
        nbands, data_shape, interpolators = cls._setup_interpolators(
            data, grid_kpoints, mesh_dim, sort_idx, gaussian
        )
        return cls(nbands, data_shape, interpolators, ncl, gpoints)

    @classmethod
    def from_file(cls, filename):
        coeff, gpoints, kpoints, structure = load_coefficients(filename)
        return cls.from_coefficients(coeff, gpoints, kpoints, structure)

    @classmethod
    def from_coefficients(
        cls, coefficients, gpoints, kpoints, structure, symprec=defaults["symprec"]
    ):
        logger.info("Initializing wavefunction overlap calculator")

        mesh_dim = get_mesh_from_kpoint_numbers(kpoints)
        if np.product(mesh_dim) == len(kpoints):
            return cls.from_data(kpoints, coefficients, gpoints)

        full_kpoints, *symmetry_mapping = expand_kpoints(
            structure, kpoints, time_reversal=True, return_mapping=True, symprec=symprec
        )
        coefficients = desymmetrize_coefficients(
            coefficients, gpoints, kpoints, structure, *symmetry_mapping
        )
        return cls.from_data(full_kpoints, coefficients, gpoints)

    def get_coefficients(self, spin, bands, kpoints):
        interp_coeffs = self.interpolate(spin, bands, kpoints)
        if self.ncl:
            interp_coeffs /= np.linalg.norm(interp_coeffs, axis=(-2, -1))[
                ..., None, None
            ]
        else:
            interp_coeffs /= np.linalg.norm(interp_coeffs, axis=-1)[:, None]
        return interp_coeffs

    def get_overlap(self, spin, band_a, kpoint_a, band_b, kpoint_b):
        bands, kpoints, single_overlap = group_bands_and_kpoints(
            band_a, kpoint_a, band_b, kpoint_b
        )
        p = self.get_coefficients(spin, bands, kpoints)
        overlap = get_overlap(p[0], p[1:])
        if single_overlap:
            return overlap[0]
        else:
            return overlap



"""
This module contains a script for using amset from the command line.
"""
import warnings

import click
from ruamel.yaml.error import MantissaNoDotYAML1_1Warning

from amset.tools.deformation import deform
from amset.tools.effmass import eff_mass
from amset.tools.phonon_frequency import phonon_frequency
from amset.tools.plot import plot
from amset.tools.run import run
from amset.tools.wavefunction import wave

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

warnings.filterwarnings("ignore", category=UserWarning, module="pymatgen")
warnings.filterwarnings("ignore", category=RuntimeWarning, module="amset")
warnings.filterwarnings("ignore", category=FutureWarning, module="scipy")
warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.simplefilter("ignore", MantissaNoDotYAML1_1Warning)


def safe_cli():
    try:
        cli()
    except Exception as e:
        click.echo(str(e))


@click.group(context_settings=dict(help_option_names=["-h", "--help"]))
def cli():
    """
    AMSET is a tool to calculate carrier transport properties from ab initio
    calculation data
    """
    pass


cli.add_command(plot)
cli.add_command(run)
cli.add_command(phonon_frequency)
cli.add_command(wave)
cli.add_command(eff_mass)
cli.add_command(deform)



from pathlib import Path

import pytest
from click.testing import CliRunner
from pymatgen import Spin

from amset.deformation.io import load_deformation_potentials
from amset.tools.deformation import read


@pytest.mark.parametrize(
    "options,nbands",
    [
        pytest.param([], 5, id="default"),
        pytest.param(["--bands", "2:7"], 6, id="bands"),
    ],
)
def test_read(clean_dir, test_dir, options, nbands):
    deform_dir = test_dir / "deformation"
    folders = [str(deform_dir / f"0{i}") for i in range(4)]

    runner = CliRunner()
    result = runner.invoke(read, folders + options)

    assert result.exit_code == 0
    assert "band:    4  k-point: [ -0.38  -0.38   0.00  ]" in result.output
    assert "[  8.71   0.42   0.00]" in result.output

    deformation_file = Path("deformation.h5")
    assert deformation_file.exists()

    deformation_potentials, _, _ = load_deformation_potentials(deformation_file)
    assert deformation_potentials[Spin.up][4, 0, 0, 0] == pytest.approx(3.25)
    assert deformation_potentials[Spin.up][3, 10, 0, 0] == pytest.approx(7.3)

    assert len(deformation_potentials[Spin.up]) == nbands



"""Module defining constants and default parameters."""
import numpy as np
from monty.serialization import loadfn
from pkg_resources import resource_filename
from pymatgen import Spin
from scipy.constants import physical_constants

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

# distance
bohr_to_m = physical_constants["Bohr radius"][0]
bohr_to_cm = bohr_to_m * 100
bohr_to_nm = bohr_to_m * 1e9
bohr_to_angstrom = bohr_to_m * 1e10
m_to_bohr = 1 / bohr_to_m
cm_to_bohr = 1 / (bohr_to_m * 100)
nm_to_bohr = 1 / (bohr_to_m * 1e9)
angstrom_to_bohr = 1 / bohr_to_angstrom

# energy
hartree_to_joule = physical_constants["Hartree energy"][0]
hartree_to_ev = physical_constants["Hartree energy in eV"][0]
joule_to_hartree = 1 / hartree_to_joule
ev_to_hartree = 1 / hartree_to_ev

# time
au_to_s = physical_constants["atomic unit of time"][0]
s_to_au = 1 / au_to_s

# current
e_si = physical_constants["elementary charge"][0]
ampere_to_au = 1 / physical_constants["atomic unit of current"][0]
coulomb_to_au = ampere_to_au * s_to_au

# misc
boltzmann_si = physical_constants["Boltzmann constant"][0]
boltzmann_au = boltzmann_si * joule_to_hartree
hbar = physical_constants["Planck constant over 2 pi in eV s"][0]
gpa_to_au = joule_to_hartree * 1e9 / m_to_bohr ** 3

ktol = 1e-5
small_val = 1e-32  # e.g. used for an initial non-zero val
output_width = 69
numeric_types = (float, int, np.integer, np.floating)

spin_name = {Spin.up: "spin-up", Spin.down: "spin-down"}
str_to_spin = {"up": Spin.up, "down": Spin.down}
spin_to_int = {Spin.up: 0, Spin.down: 1}
int_to_spin = {0: Spin.up, 1: Spin.down}

defaults = loadfn(resource_filename("amset", "defaults.yaml"))



import click

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

zero_weighted_type = click.Choice(["keep", "drop", "prefer"], case_sensitive=False)
image_type = click.Choice(["pdf", "png", "svg"], case_sensitive=False)
path_type = click.Path(exists=True)


def echo_ibands(ibands, is_spin_polarized):
    if is_spin_polarized:
        click.echo("Including:")
        for spin, spin_bands in ibands.items():
            min_b = spin_bands.min() + 1
            max_b = spin_bands.max() + 1
            click.echo("  - Spin-{} bands {}—{}".format(spin.name, min_b, max_b))
    else:
        bands = list(ibands.values())[0]
        click.echo("Including bands {}—{}".format(bands.min() + 1, bands.max() + 1))



import numpy as np
import pytest
from pymatgen.io.ase import AseAtomsAdaptor
from pytest import mark
from spglib import get_ir_reciprocal_mesh

from amset.electronic_structure.symmetry import (
    expand_kpoints,
    get_rotation_angle,
    get_rotation_axis,
    get_symmetry_type,
    rotation_matrix_to_su2,
)


@mark.parametrize(
    "symmetry_matrix,expected",
    [
        pytest.param([[1, 0, 0], [0, 1, 0], [0, 0, 1]], 1, id="identity"),
        pytest.param([[-1, 0, 0], [0, -1, 0], [0, 0, -1]], 2, id="inversion"),
        pytest.param([[1, 0, 0], [0, 0, -1], [0, 1, 0]], 3, id="x-rotation (90)"),
        pytest.param([[1, 0, 0], [0, -1, 0], [0, 0, -1]], 4, id="x-rotation (180)"),
        pytest.param(
            [
                [0.5, 0.5, 0.7071068],
                [0.5, 0.5, -0.7071068],
                [-0.7071068, 0.7071068, 0.0],
            ],
            3,
            id="xy-rotation (90)",
        ),
        pytest.param([[0, 1, 0], [1, 0, 0], [0, 0, -1]], 4, id="xy-rotation (180)"),
        pytest.param([[-1, 0, 0], [0, 1, 0], [0, 0, 1]], 5, id="x-reflection"),
        pytest.param([[0, -1, 0], [-1, 0, 0], [0, 0, 1]], 5, id="xy-reflection"),
        pytest.param(
            [[-1, 0, 0], [0, 0, -1], [0, 1, 0]], 6, id="x-improper rotation (90)"
        ),
        pytest.param(
            [[0, -1, 0], [1, 0, 0], [-1, 0, 0]],
            pytest.raises(ValueError),
            id="malformed matrix",
        ),
    ],
)
def test_get_symmetry_type(symmetry_matrix, expected):
    if not isinstance(expected, int):
        with expected:
            get_symmetry_type(symmetry_matrix)
    else:
        assert get_symmetry_type(symmetry_matrix) == expected


@pytest.mark.parametrize(
    "symmetry_matrix,expected",
    [
        pytest.param(
            [[1, 0, 0], [0, 1, 0], [0, 0, 1]], pytest.raises(ValueError), id="identity"
        ),
        pytest.param(
            [[-1, 0, 0], [0, -1, 0], [0, 0, -1]],
            pytest.raises(ValueError),
            id="inversion",
        ),
        pytest.param([[1, 0, 0], [0, 0, -1], [0, 1, 0]], 90, id="x-rotation (90)"),
        pytest.param([[1, 0, 0], [0, -1, 0], [0, 0, -1]], 180, id="x-rotation (180)"),
        pytest.param(
            [
                [0.5, 0.5, 0.7071068],
                [0.5, 0.5, -0.7071068],
                [-0.7071068, 0.7071068, 0.0],
            ],
            90,
            id="xy-rotation (90)",
        ),
        pytest.param([[0, 1, 0], [1, 0, 0], [0, 0, -1]], 180, id="xy-rotation (180)"),
        pytest.param(
            [[-1, 0, 0], [0, 1, 0], [0, 0, 1]],
            pytest.raises(ValueError),
            id="x-reflection",
        ),
        pytest.param(
            [[0, -1, 0], [-1, 0, 0], [0, 0, 1]],
            pytest.raises(ValueError),
            id="xy-reflection",
        ),
        pytest.param(
            [[-1, 0, 0], [0, 0, -1], [0, 1, 0]], 90, id="x-improper rotation (90)"
        ),
        pytest.param(
            [[0, -1, 0], [1, 0, 0], [-1, 0, 0]],
            pytest.raises(ValueError),
            id="malformed matrix",
        ),
    ],
)
def test_get_rotation_angle(symmetry_matrix, expected):
    if not isinstance(expected, int):
        with expected:
            get_rotation_angle(symmetry_matrix)
    else:
        assert get_rotation_angle(symmetry_matrix) == expected


@pytest.mark.parametrize(
    "symmetry_matrix,expected",
    [
        pytest.param(
            [[1, 0, 0], [0, 1, 0], [0, 0, 1]], pytest.raises(ValueError), id="identity"
        ),
        pytest.param(
            [[-1, 0, 0], [0, -1, 0], [0, 0, -1]],
            pytest.raises(ValueError),
            id="inversion",
        ),
        pytest.param(
            [[1, 0, 0], [0, 0, -1], [0, 1, 0]], [1, 0, 0], id="x-rotation (90)"
        ),
        pytest.param(
            [[1, 0, 0], [0, -1, 0], [0, 0, -1]], [1, 0, 0], id="x-rotation (180)"
        ),
        pytest.param(
            [
                [0.5, 0.5, 0.7071068],
                [0.5, 0.5, -0.7071068],
                [-0.7071068, 0.7071068, 0.0],
            ],
            [0.70710678118, 0.70710678118, 0.0],
            id="xy-rotation (90)",
        ),
        pytest.param(
            [[0, 1, 0], [1, 0, 0], [0, 0, -1]],
            [0.70710678118, 0.70710678118, 0.0],
            id="xy-rotation (180)",
        ),
        pytest.param(
            [[-1, 0, 0], [0, 1, 0], [0, 0, 1]],
            pytest.raises(ValueError),
            id="x-reflection",
        ),
        pytest.param(
            [[0, -1, 0], [-1, 0, 0], [0, 0, 1]],
            pytest.raises(ValueError),
            id="xy-reflection",
        ),
        pytest.param(
            [[-1, 0, 0], [0, 0, -1], [0, 1, 0]],
            [1, 0, 0],
            id="x-improper rotation (90)",
        ),
        pytest.param(
            [[0, -1, 0], [1, 0, 0], [-1, 0, 0]],
            pytest.raises(ValueError),
            id="malformed matrix",
        ),
    ],
)
def test_get_rotation_axis(symmetry_matrix, expected):
    if not isinstance(expected, list):
        with expected:
            get_rotation_axis(symmetry_matrix)
    else:
        axis = get_rotation_axis(symmetry_matrix)
        np.testing.assert_array_almost_equal(axis, expected)


@pytest.mark.parametrize(
    "symmetry_matrix,expected",
    [
        pytest.param(
            [[1, 0, 0], [0, 1, 0], [0, 0, 1]],
            [[1 + 0j, 0j], [0j, 1 + 0j]],
            id="identity",
        ),
        pytest.param(
            [[-1, 0, 0], [0, -1, 0], [0, 0, -1]],
            [[1 + 0j, 0j], [0j, 1 + 0j]],
            id="inversion",
        ),
        pytest.param(
            [[1, 0, 0], [0, 0, -1], [0, 1, 0]],
            [[0.707107 + 0j, -0 - 0.707107j], [0.707107j, 0.707107 + 0j]],
            id="x-rotation (90)",
        ),
        pytest.param(
            [[1, 0, 0], [0, -1, 0], [0, 0, -1]],
            [[0j, -1j], [1j, 0j]],
            id="x-rotation (180)",
        ),
        pytest.param(
            [
                [0.5, 0.5, 0.7071068],
                [0.5, 0.5, -0.7071068],
                [-0.7071068, 0.7071068, 0.0],
            ],
            [[(0.707107 + 0j), (-0.5 - 0.5j)], [(-0.5 + 0.5j), (0.707107 + 0j)]],
            id="xy-rotation (90)",
        ),
        pytest.param(
            [[0, 1, 0], [1, 0, 0], [0, 0, -1]],
            [[0j, -0.7071068 - 0.707107j], [-0.7071068 + 0.707107j, 0j]],
            id="xy-rotation (180)",
        ),
        pytest.param(
            [[-1, 0, 0], [0, 1, 0], [0, 0, 1]], [[0j, -1j], [1j, 0j]], id="x-reflection"
        ),
        pytest.param(
            [[0, -1, 0], [-1, 0, 0], [0, 0, 1]],
            [[0j, -0.7071068 - 0.707107j], [-0.7071068 + 0.707107j, 0j]],
            id="xy-reflection",
        ),
        pytest.param(
            [[-1, 0, 0], [0, 0, -1], [0, 1, 0]],
            [[0.707107 + 0j, 0.707107j], [-0 - 0.707107j, 0.707107 + 0j]],
            id="x-improper rotation (90)",
        ),
        pytest.param(
            [[0, -1, 0], [1, 0, 0], [-1, 0, 0]],
            pytest.raises(ValueError),
            id="malformed matrix",
        ),
    ],
)
def test_rotation_matrix_to_su2(symmetry_matrix, expected):
    if not isinstance(expected, list):
        with expected:
            rotation_matrix_to_su2(symmetry_matrix)
    else:
        su2 = rotation_matrix_to_su2(symmetry_matrix)
        np.testing.assert_array_almost_equal(su2, expected)


@pytest.fixture(params=[0, 1], ids=["gamma", "shift"])
def shift(request):
    idx = request.param
    return np.array([idx, idx, idx])


@pytest.fixture(
    params=[
        [5, 5, 5],
        [6, 6, 6],
        [21, 21, 21],
        [22, 22, 22],
        [5, 7, 9],
        [6, 8, 10],
        [21, 8, 13],
        [4, 5, 1],
    ],
    ids=[
        "odd (iso)",
        "even (iso)",
        "large odd (iso)",
        "large even (iso)",
        "odd (aniso)",
        "even (aniso)",
        "mixed (aniso)",
        "2D material",
    ],
)
def mesh(request):
    return np.array(request.param)


def test_expand_kpoints(symmetry_structure, shift, mesh):
    def _kpoints_to_first_bz(kp):
        """helper function to map k-points to 1st BZ"""
        kp = np.array(kp)
        kp = kp - np.round(kp)
        kp[kp.round(8) == -0.5] = 0.5
        return kp

    def _sort_kpoints(kp):
        """Helper function to put k-points in a consistent order"""
        kp = kp.round(8)
        sort_idx = np.lexsort((kp[:, 2], kp[:, 1], kp[:, 0]))
        return kp[sort_idx]

    # generate true k-points and IR k-points using spglib
    atoms = AseAtomsAdaptor.get_atoms(symmetry_structure)
    mapping, addresses = get_ir_reciprocal_mesh(mesh, atoms, is_shift=shift)
    true_kpoints = addresses / mesh + shift / (mesh * 2)
    true_kpoints = _kpoints_to_first_bz(true_kpoints)
    true_kpoints_sort = _sort_kpoints(true_kpoints)

    ir_mapping = np.unique(mapping, return_index=False)
    ir_kpoints = true_kpoints[ir_mapping]

    # try to expand the irreducible k-points back  to the full BZ
    full_kpoints, rots, _, _, op_mapping, kp_mapping = expand_kpoints(
        symmetry_structure, ir_kpoints, return_mapping=True
    )
    full_kpoints = _kpoints_to_first_bz(full_kpoints)
    full_kpoints_sort = _sort_kpoints(full_kpoints)

    # assert final k-points match the expected true k-points
    diff = np.linalg.norm(full_kpoints_sort - true_kpoints_sort, axis=1)
    assert np.max(diff) == 0

    # now ensure that the rotation mapping actually works
    rotated_kpoints = []
    for r, k in zip(op_mapping, kp_mapping):
        rotated_kpoints.append(np.dot(rots[r], ir_kpoints[k]))
    rotated_kpoints = _kpoints_to_first_bz(rotated_kpoints)
    rotated_kpoints_sort = _sort_kpoints(rotated_kpoints)

    # assert rotated k-points match the expected true k-points
    diff = np.linalg.norm(rotated_kpoints_sort - true_kpoints_sort, axis=1)
    assert np.max(diff) == 0



import logging

import numpy as np
from pymatgen.analysis.elasticity.strain import Strain
from pymatgen.transformations.standard_transformations import (
    DeformStructureTransformation,
)

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

logger = logging.getLogger(__name__)


def get_strain_fields():
    """Generates a list of strain fields for a rank 2 strain tensor in Voigt notation"""
    inds = [(i,) for i in range(6)]
    strain_fields = np.zeros((len(inds), 6))
    for n, i in enumerate(inds):
        np.put(strain_fields[n], i, 1)
    strain_fields[:, 3:] *= 2
    return strain_fields


def get_strains(distance=0.005):
    strain_fields = get_strain_fields()
    strains = []
    for strain_field in strain_fields:
        strains.append(Strain.from_voigt(strain_field * abs(distance)))
        strains.append(Strain.from_voigt(strain_field * -abs(distance)))
    return strains


def get_deformations(distance):
    strains = get_strains(distance=distance)
    deformations = [s.get_deformation_matrix() for s in strains]
    return deformations


def get_deformed_structures(structure, deformations):
    deformed_structures = []
    for i, deformation in enumerate(deformations):
        dst = DeformStructureTransformation(deformation)
        deformed_structures.append(dst.apply_transformation(structure))
    return deformed_structures



import logging

import numpy as np

from amset.constants import defaults
from amset.deformation.common import desymmetrize_deformation_potentials
from amset.deformation.io import load_deformation_potentials
from amset.electronic_structure.kpoints import get_mesh_from_kpoint_numbers
from amset.electronic_structure.symmetry import expand_kpoints
from amset.interpolation.periodic import PeriodicLinearInterpolator

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

logger = logging.getLogger(__name__)


class DeformationPotentialInterpolator(PeriodicLinearInterpolator):
    @classmethod
    def from_file(cls, filename, scale=1.0):
        deform_potentials, kpoints, structure = load_deformation_potentials(filename)
        deform_potentials = {s: d * scale for s, d in deform_potentials.items()}
        return cls.from_deformation_potentials(deform_potentials, kpoints, structure)

    @classmethod
    def from_deformation_potentials(
        cls, deformation_potentials, kpoints, structure, symprec=defaults["symprec"]
    ):
        logger.info("Initializing deformation potential interpolator")

        mesh_dim = get_mesh_from_kpoint_numbers(kpoints)
        if np.product(mesh_dim) == len(kpoints):
            return cls.from_data(kpoints, deformation_potentials)

        full_kpoints, rotations, _, _, op_mapping, kp_mapping = expand_kpoints(
            structure, kpoints, time_reversal=True, return_mapping=True, symprec=symprec
        )
        logger.warning("Desymmetrizing deformation potentials, this could go wrong.")
        deformation_potentials = desymmetrize_deformation_potentials(
            deformation_potentials, structure, rotations, op_mapping, kp_mapping
        )
        return cls.from_data(full_kpoints, deformation_potentials)



import logging
import time
from typing import Dict, List, Union

import numpy as np
from pymatgen import Spin

from amset.constants import int_to_spin, numeric_types
from amset.electronic_structure.symmetry import (
    rotation_matrix_to_cartesian,
    rotation_matrix_to_su2,
)
from amset.log import log_time_taken
from amset.util import get_progress_bar

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

logger = logging.getLogger(__name__)


def sample_random_kpoints(
    nspins: int, ikpoints: int, nband: Dict[Spin, Union[List[int], int]], n_samples: int
):
    spin_idxs = []
    band_idxs = []
    kpoint_idxs = []

    for spin_idx in range(nspins):
        spin = int_to_spin[spin_idx]

        spin_iband = nband[spin]
        if isinstance(spin_iband, numeric_types):
            spin_iband = np.arange(spin_iband, dtype=int)

        s_idxs = [spin_idx] * n_samples
        b_idxs = np.random.randint(min(spin_iband), max(spin_iband) + 1, n_samples)
        k_idxs = np.random.choice(ikpoints, n_samples)

        spin_idxs.append(s_idxs)
        band_idxs.append(b_idxs)
        kpoint_idxs.append(k_idxs)

    spin_idxs = np.concatenate(spin_idxs)
    band_idxs = np.concatenate(band_idxs)
    kpoint_idxs = np.concatenate(kpoint_idxs)
    return np.stack([spin_idxs, band_idxs, kpoint_idxs], axis=1)


def desymmetrize_coefficients(
    coeffs,
    gpoints,
    kpoints,
    structure,
    rotations,
    translations,
    is_tr,
    op_mapping,
    kp_mapping,
    pbar=True,
):
    logger.info("Desymmetrizing wavefunction coefficients")
    t0 = time.perf_counter()

    ncl = is_ncl(coeffs)
    rots = rotations[op_mapping]
    taus = translations[op_mapping]
    trs = is_tr[op_mapping]

    su2s = None
    if ncl:
        # get cartesian rotation matrix
        r_cart = [rotation_matrix_to_cartesian(r, structure.lattice) for r in rotations]

        # calculate SU(2)
        su2_no_dagger = np.array([rotation_matrix_to_su2(r) for r in r_cart])

        # calculate SU(2)^{dagger}
        su2 = np.conjugate(su2_no_dagger).transpose((0, 2, 1))
        su2s = su2[op_mapping]

    g_mesh = (np.abs(gpoints).max(axis=0) + 3) * 2
    g1, g2, g3 = (gpoints + g_mesh / 2).astype(int).T  # indices of g-points to keep

    all_rot_coeffs = {}
    for spin, spin_coeffs in coeffs.items():
        coeff_shape = (len(spin_coeffs), len(rots)) + tuple(g_mesh)
        if ncl:
            coeff_shape += (2,)
        rot_coeffs = np.zeros(coeff_shape, dtype=complex)

        state_idxs = list(range(len(rots)))
        if pbar:
            state_idxs = get_progress_bar(state_idxs, desc="progress")

        for k_idx in state_idxs:
            map_idx = kp_mapping[k_idx]

            rot = rots[k_idx]
            tau = taus[k_idx]
            tr = trs[k_idx]
            kpoint = kpoints[map_idx]

            rot_kpoint = np.dot(rot, kpoint)
            kdiff = np.around(rot_kpoint)
            rot_kpoint -= kdiff

            edges = np.around(rot_kpoint, 5) == -0.5
            rot_kpoint += edges
            kdiff -= edges

            rot_gpoints = np.dot(rot, gpoints.T).T
            rot_gpoints = np.around(rot_gpoints).astype(int)
            rot_gpoints += kdiff.astype(int)

            if tr:
                tau = -tau

            factor = np.exp(-1j * 2 * np.pi * np.dot(rot_gpoints + rot_kpoint, tau))
            rg1, rg2, rg3 = (rot_gpoints + g_mesh / 2).astype(int).T

            if ncl:
                # perform rotation in spin space
                su2 = su2s[k_idx]
                rc = np.zeros_like(spin_coeffs[:, map_idx])
                rc[:, :, 0] = (
                    su2[0, 0] * spin_coeffs[:, map_idx, :, 0]
                    + su2[0, 1] * spin_coeffs[:, map_idx, :, 1]
                )
                rc[:, :, 1] = (
                    su2[1, 0] * spin_coeffs[:, map_idx, :, 0]
                    + su2[1, 1] * spin_coeffs[:, map_idx, :, 1]
                )
                rot_coeffs[:, k_idx, rg1, rg2, rg3] = factor[None, :, None] * rc
            else:
                rot_coeffs[:, k_idx, rg1, rg2, rg3] = spin_coeffs[:, map_idx] * factor

            if tr and not ncl:
                rot_coeffs[:, k_idx] = np.conjugate(rot_coeffs[:, k_idx])

        all_rot_coeffs[spin] = rot_coeffs[:, :, g1, g2, g3]

    log_time_taken(t0)
    return all_rot_coeffs


def get_gpoints(reciprocal_lattice, nbmax, encut, kpoint=(0, 0, 0)):
    all_g = np.array(list(np.ndindex(tuple(2 * nbmax[::-1] + 1))))
    all_g = all_g[:, [2, 1, 0]]  # swap columns

    all_g[all_g[:, 2] > nbmax[2], 2] -= 2 * nbmax[2] + 1
    all_g[all_g[:, 1] > nbmax[1], 1] -= 2 * nbmax[1] + 1
    all_g[all_g[:, 0] > nbmax[0], 0] -= 2 * nbmax[0] + 1

    cart_g = np.dot(all_g + kpoint, reciprocal_lattice)
    norm_g = np.linalg.norm(cart_g, axis=1)
    ener_g = norm_g ** 2 / 0.262465831

    return all_g[ener_g <= encut]


def get_min_gpoints(nbmax):
    min_gpoint = -nbmax
    num_gpoint = nbmax * 2 + 1
    return min_gpoint, num_gpoint


def get_gpoint_indices(gpoints, min_gpoint, num_gpoint):
    shifted_g = gpoints - min_gpoint
    nyz = num_gpoint[1] * num_gpoint[2]
    nz = num_gpoint[2]
    indices = shifted_g[:, 0] * nyz + shifted_g[:, 1] * nz + shifted_g[:, 2]
    return indices.astype(int)


def is_ncl(coefficients):
    return len(list(coefficients.values())[0].shape) == 4


def get_overlap(origin, final):
    if len(origin.shape) == 2:
        # ncl
        return (
            np.abs(
                np.dot(np.conj(origin[:, 0]), final[:, :, 0].T)
                + np.dot(np.conj(origin[:, 1]), final[:, :, 1].T)
            )
            ** 2
        )
    else:
        return np.abs(np.dot(np.conj(origin), final.T)) ** 2



"""
Customised implementation of FermiDos. Will Move back to pymatgen at some point.
"""
import logging
from typing import Dict, Optional, Tuple, Union

import numpy as np
from monty.json import MSONable
from pymatgen import Spin, Structure
from pymatgen.electronic_structure.dos import Dos

from amset.constants import boltzmann_au, ev_to_hartree, hartree_to_ev
from amset.electronic_structure.fd import fd

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

logger = logging.getLogger(__name__)


class FermiDos(Dos, MSONable):
    """
    This wrapper class helps relate the density of states, doping levels
    (i.e. carrier concentrations) and corresponding fermi levels. A negative
    doping concentration indicates the majority carriers are electrons
    (n-type doping); a positive doping concentration indicates holes are the
    majority carriers (p-type doping).

    Args:
        efermi: The Fermi level energy in Hartree.
        energies: A sequences of energies in Hartree.
        densities ({Spin: np.array}): representing the density of states
            for each Spin.
        structure: A structure. If not provided, the structure
            of the dos object will be used. If the dos does not have an
            associated structure object, an error will be thrown.
        dos_weight: The weighting for the dos. Defaults to 2 for non-spin
            polarized calculations and 1 for spin-polarized calculations.
        atomic_units: Whether energies are given in eV or Hartree.
        num_electrons: The number of electrons in the system. If None, this will be
            calculated by integrating up to the intrinsic Fermi level.
    """

    def __init__(
        self,
        efermi: float,
        energies: np.ndarray,
        densities: Dict[Spin, np.ndarray],
        structure: Structure,
        dos_weight: Optional[float] = None,
        atomic_units: bool = True,
        num_electrons: Optional[float] = None,
    ):
        # structure should be atomic structure
        super().__init__(efermi, energies, densities)
        self.structure = structure
        self.atomic_units = atomic_units

        if not dos_weight:
            dos_weight = 2 if len(self.densities) == 1 else 1

        self.dos_weight = dos_weight
        self.tdos = np.array(self.get_densities()) * self.dos_weight
        self.de = self.energies[1] - self.energies[0]
        self._num_electrons = num_electrons  # this is just for msonability

        if num_electrons is None:
            # integrate up to Fermi level to get number of electrons
            self.nelect = self.tdos[self.energies <= self.efermi].sum() * self.de
        else:
            self.nelect = num_electrons

        logger.info(
            "Intrinsic DOS Fermi level: {:.4f} eV".format(
                self.efermi * hartree_to_ev if atomic_units else self.efermi
            )
        )
        logger.info("DOS contains {:.3f} electrons".format(self.nelect))

    def get_doping(
        self,
        fermi_level: float,
        temperature: float,
        return_electron_hole_conc: bool = False,
    ) -> Union[float, Tuple[float, float, float]]:
        """
        Calculate the doping (majority carrier concentration) at a given
        fermi level  and temperature. A simple Left Riemann sum is used for
        integrating the density of states over energy & equilibrium Fermi-Dirac
        distribution.

        Args:
            fermi_level: The fermi_level level in Hartree.
            temperature: The temperature in Kelvin.
            return_electron_hole_conc: Whether to also return the separate
                electron and hole concentrations at the doping level.

        Returns:
            If return_electron_hole_conc is False: the doping concentration in
            units of 1/Bohr^3. Negative values indicate that the majority carriers
            are electrons (n-type doping) whereas positive values indicates the
            majority carriers are holes (p-type doping).

            If return_electron_hole_conc is True: the doping concentration,
            electron concentration and hole concentration as a tuple.
        """
        wdos = _get_weighted_dos(
            self.energies,
            self.tdos,
            fermi_level,
            temperature,
            atomic_units=self.atomic_units,
        )

        num_electrons = wdos.sum() * self.de
        conc = (self.nelect - num_electrons) / self.structure.volume

        if return_electron_hole_conc:
            cb_conc = wdos[self.energies > self.efermi].sum() * self.de
            vb_conc = wdos[self.energies <= self.efermi].sum() * self.de
            cb_conc = cb_conc / self.structure.volume
            vb_conc = (self.nelect - vb_conc) / self.structure.volume
            return conc, cb_conc, vb_conc

        else:
            return conc

    def get_num_electrons(self, fermi_level: float, temperature: float) -> float:
        """
        Calculate the number of electrons at a given fermi level and temperature.
        A simple Left Riemann sum is used for integrating the density of states over
        energy & equilibrium Fermi-Dirac distribution.

        Args:
            fermi_level: The fermi_level level in Hartree.
            temperature: The temperature in Kelvin.

        Returns:
            The number of electrons.
        """
        wdos = _get_weighted_dos(
            self.energies,
            self.tdos,
            fermi_level,
            temperature,
            atomic_units=self.atomic_units,
        )

        num_electrons = wdos.sum() * self.de
        return num_electrons

    def get_fermi_from_num_electrons(
        self,
        num_electrons: float,
        temperature: float,
        tol: float = 0.01,
        nstep: int = 50,
        step: float = 0.1,
        precision: int = 10,
    ):
        # this is finding the Fermi level of metals
        fermi = self.efermi  # initialize target fermi
        relative_error = float("inf")
        for _ in range(precision):
            frange = np.arange(-nstep, nstep + 1) * step + fermi
            calc_nelectrons = [self.get_num_electrons(f, temperature) for f in frange]
            relative_error = abs(np.array(calc_nelectrons) / num_electrons - 1.0)
            fermi = frange[np.argmin(relative_error)]
            step /= 10.0

        if min(relative_error) > tol:
            raise ValueError(
                "Could not find fermi within {}% of num electrons={}".format(
                    tol * 100, num_electrons
                )
            )

        return fermi

    def get_fermi(
        self,
        concentration: float,
        temperature: float,
        tol: float = 0.01,
        nstep: int = 50,
        step: float = 0.1,
        precision: int = 10,
        return_electron_hole_conc=False,
    ):
        """
        Finds the fermi level at which the doping concentration at the given
        temperature (T) is equal to concentration. A greedy algorithm is used
        where the relative error is minimized by calculating the doping at a
        grid which continually becomes finer.

        Args:
            concentration: The doping concentration in 1/Bohr^3. Negative values
                represent n-type doping and positive values represent p-type
                doping.
            temperature: The temperature in Kelvin.
            return_electron_hole_conc: Whether to also return the separate
                electron and hole concentrations at the doping level.

        Returns:
            If return_electron_hole_conc is False: The Fermi level in eV. Note
            that this is different from the default dos.efermi.

            If return_electron_hole_conc is True: the Fermi level, electron
            concentration and hole concentration at the Fermi level as a tuple.
            The electron and hole concentrations are in Bohr^-3.
        """
        fermi = self.efermi  # initialize target fermi
        relative_error = float("inf")
        for _ in range(precision):
            frange = np.arange(-nstep, nstep + 1) * step + fermi
            calc_doping = np.array([self.get_doping(f, temperature) for f in frange])
            relative_error = abs(calc_doping / concentration - 1.0)
            fermi = frange[np.argmin(relative_error)]
            step /= 10.0

        if min(relative_error) > tol:
            raise ValueError(
                "Could not find fermi within {}% of concentration={}".format(
                    tol * 100, concentration
                )
            )

        if return_electron_hole_conc:
            _, n_elec, n_hole = self.get_doping(
                fermi, temperature, return_electron_hole_conc=True
            )
            return fermi, n_elec, n_hole
        else:
            return fermi


def _get_weighted_dos(energies, dos, fermi_level, temperature, atomic_units=True):
    if temperature == 0.0:
        occ = np.where(energies < fermi_level, 1.0, 0.0)
        occ[energies == fermi_level] = 0.5
    else:
        kbt = temperature * boltzmann_au
        if atomic_units:
            occ = fd(energies, fermi_level, kbt)
        else:
            occ = fd(energies * ev_to_hartree, fermi_level * ev_to_hartree, kbt)

    wdos = dos * occ
    return wdos



import os
import unittest

import numpy as np
from numpy.testing import assert_array_equal
from pymatgen import Spin
from pymatgen.io.vasp import Vasprun

from amset.interpolation.bandstructure import Interpolator
from amset.log import initialize_amset_logger

test_dir = os.path.dirname(os.path.abspath(__file__))
gaas_files = os.path.join(test_dir, "..", "..", "..", "examples", "GaAs")
tin_dioxide_files = os.path.join(test_dir, "..", "..", "..", "examples", "SnO2")
si_files = os.path.join(test_dir, "..", "..", "..", "examples", "Si")
pbs_files = os.path.join(test_dir, "..", "..", "..", "examples", "PbS")


@unittest.skip("Outdated tests")
class TestBoltzTraP2Interpolater(unittest.TestCase):
    """Tests for interpolating a band structure using BoltzTraP2."""

    def setUp(self):
        vr = Vasprun(
            os.path.join(gaas_files, "vasprun.xml.gz"), parse_projected_eigen=True
        )
        bs = vr.get_band_structure()
        num_electrons = vr.parameters["NELECT"]

        self.kpoints = np.array(vr.actual_kpoints)
        self.interpolater = Interpolator(
            bs, num_electrons, interpolate_projections=True, interpolation_factor=1
        )

    def test_initialisation(self):
        """Test coefficients and parameters are calculated correctly."""
        self.interpolater.initialize()
        params = self.interpolater._parameters
        assert_array_equal(params[0][0], [[0, 0, 0]])
        self.assertAlmostEqual(params[1][0][0], 9.14055614)
        self.assertAlmostEqual(params[2][0][0].real, -2.33144546)

    def test_get_energies(self):
        """Test getting the interpolated energy, velocity and effective mass."""

        # test just getting energy
        energies = self.interpolater.get_energies(
            self.kpoints, 25, return_velocity=False, curvature=False
        )
        self.assertEqual(energies.shape, (138,))
        self.assertAlmostEqual(energies[0], 3.852399483908641)

        # test energy + velocity
        energies, velocities = self.interpolater.get_energies(
            self.kpoints, 25, return_velocity=True, curvature=False
        )
        self.assertEqual(energies.shape, (138,))
        self.assertAlmostEqual(energies[0], 3.852399483908641)
        self.assertEqual(velocities.shape, (138, 3))
        self.assertAlmostEqual(velocities[10][0], 5.401166156893334e6, places=0)

        # test energy + effective_mass
        energies, effective_masses = self.interpolater.get_energies(
            self.kpoints, 25, return_velocity=False, curvature=True
        )
        self.assertEqual(energies.shape, (138,))
        self.assertAlmostEqual(energies[0], 3.852399483908641)
        self.assertEqual(effective_masses.shape, (138, 3, 3))
        self.assertAlmostEqual(effective_masses[10][0][0], 0.2242569613626494)

        # test energy + velocity + effective_mass
        energies, velocities, effective_masses = self.interpolater.get_energies(
            self.kpoints, 25, return_velocity=True, curvature=True
        )
        self.assertEqual(energies.shape, (138,))
        self.assertAlmostEqual(energies[0], 3.852399483908641)
        self.assertEqual(velocities.shape, (138, 3))
        self.assertAlmostEqual(velocities[10][0], 5.401166156893334e6, places=0)
        self.assertEqual(effective_masses.shape, (138, 3, 3))
        self.assertAlmostEqual(effective_masses[10][0][0], 0.2242569613626494)

    def test_get_energies_scissor(self):
        """Test scissoring of band energies."""

        # test valence band
        energies = self.interpolater.get_energies(
            self.kpoints, 25, return_velocity=False, curvature=False, scissor=1.0
        )
        self.assertEqual(energies.shape, (138,))
        self.assertAlmostEqual(energies[0], 3.852399483908641 - 0.5)

        # test conduction band
        energies = self.interpolater.get_energies(
            self.kpoints, 33, return_velocity=False, curvature=False, scissor=1.0
        )
        self.assertEqual(energies.shape, (138,))
        self.assertAlmostEqual(energies[0], 7.301700765 + 0.5)

    def test_get_energies_multiple_bands(self):
        """Test getting the interpolated data for multiple bands."""
        # test just getting energy
        energies = self.interpolater.get_energies(
            self.kpoints, [25, 35], return_velocity=False, curvature=False
        )
        self.assertEqual(energies.shape, (2, 138))
        self.assertAlmostEqual(energies[0][0], 3.852399483908641)
        self.assertAlmostEqual(energies[1][0], 9.594401616456384)

        # test energy + velocity
        energies, velocities = self.interpolater.get_energies(
            self.kpoints, [25, 35], return_velocity=True, curvature=False
        )
        self.assertEqual(energies.shape, (2, 138))
        self.assertAlmostEqual(energies[0][0], 3.852399483908641)
        self.assertAlmostEqual(energies[1][0], 9.594401616456384)
        self.assertEqual(velocities.shape, (2, 138, 3))
        self.assertAlmostEqual(velocities[0][10][0], 5.401166156893334e6, places=0)
        self.assertAlmostEqual(velocities[1][10][0], 1.1686720831758736e8, places=0)

        # test energy + effective_mass
        energies, effective_masses = self.interpolater.get_energies(
            self.kpoints, [25, 35], return_velocity=False, curvature=True
        )
        self.assertEqual(energies.shape, (2, 138))
        self.assertAlmostEqual(energies[0][0], 3.852399483908641)
        self.assertAlmostEqual(energies[1][0], 9.594401616456384)
        self.assertEqual(effective_masses.shape, (2, 138, 3, 3))
        self.assertAlmostEqual(effective_masses[0][10][0][0], 0.224256961362649)
        self.assertAlmostEqual(effective_masses[1][10][0][0], 0.009057103344700)

        # test energy + velocity + effective_mass
        energies, velocities, effective_masses = self.interpolater.get_energies(
            self.kpoints, [25, 35], return_velocity=True, curvature=True
        )
        self.assertEqual(energies.shape, (2, 138))
        self.assertAlmostEqual(energies[0][0], 3.852399483908641)
        self.assertAlmostEqual(energies[1][0], 9.594401616456384)
        self.assertEqual(velocities.shape, (2, 138, 3))
        self.assertAlmostEqual(velocities[0][10][0], 5.401166156893334e6, places=0)
        self.assertAlmostEqual(velocities[1][10][0], 1.1686720831758736e8, places=0)
        self.assertEqual(effective_masses.shape, (2, 138, 3, 3))
        self.assertAlmostEqual(effective_masses[0][10][0][0], 0.224256961362649)
        self.assertAlmostEqual(effective_masses[1][10][0][0], 0.009057103344700)

    def test_get_energies_all_bands(self):
        # test all bands
        energies, velocities, effective_masses = self.interpolater.get_energies(
            self.kpoints, None, return_velocity=True, curvature=True
        )
        self.assertEqual(energies.shape, (96, 138))
        self.assertAlmostEqual(energies[25][0], 3.852399483908641)
        self.assertAlmostEqual(energies[35][0], 9.594401616456384)
        self.assertEqual(velocities.shape, (96, 138, 3))
        self.assertAlmostEqual(velocities[25][10][0], 5.401166156893334e6, places=0)
        self.assertAlmostEqual(velocities[35][10][0], 1.1686720831758736e8, places=0)
        self.assertEqual(effective_masses.shape, (96, 138, 3, 3))
        self.assertAlmostEqual(effective_masses[25][10][0][0], 0.22425696136264)
        self.assertAlmostEqual(effective_masses[35][10][0][0], 0.00905710334470)

    def test_get_dos(self):
        """Test generating the interpolated DOS."""
        dos = self.interpolater.get_dos([10, 10, 10], emin=-10, emax=10, width=0.075)
        self.assertEqual(dos.shape, (20000, 2))
        self.assertEqual(dos[0][0], -10)
        self.assertAlmostEqual(dos[15000][1], 3.5362612128412807)

    def test_get_energies_symprec(self):
        # vr = Vasprun(os.path.join(tin_dioxide_files, 'vasprun.xml.gz'),
        #              parse_projected_eigen=True)
        vr = Vasprun(
            os.path.join(pbs_files, "vasprun.xml.gz"), parse_projected_eigen=True
        )
        bs = vr.get_band_structure()
        num_electrons = vr.parameters["NELECT"]
        interpolater = Interpolator(
            bs, num_electrons, interpolate_projections=True, interpolation_factor=1
        )

        ir_kpoints, weights, kpoints, ir_kpoints_idx, ir_to_full_idx = get_kpoints(
            [13, 15, 29],
            vr.final_structure,
            boltztrap_ordering=True,
            return_full_kpoints=True,
        )

        initialize_amset_logger()

        (
            energies,
            velocities,
            curvature,
            projections,
            sym_info,
        ) = interpolater.get_energies(
            kpoints,
            None,
            return_velocity=True,
            atomic_units=True,
            return_curvature=True,
            return_projections=True,
            symprec=0.1,
            return_vel_outer_prod=True,
            return_kpoint_mapping=True,
        )

        (
            energies_no_sym,
            velocities_no_sym,
            curvature_no_sym,
            projections_no_sym,
        ) = interpolater.get_energies(
            kpoints,
            None,
            return_velocity=True,
            atomic_units=True,
            return_curvature=True,
            return_projections=True,
            return_vel_outer_prod=True,
            symprec=None,
        )

        np.testing.assert_array_equal(ir_to_full_idx, sym_info["ir_to_full_idx"])
        np.testing.assert_array_equal(ir_kpoints_idx, sym_info["ir_kpoints_idx"])

        # print(velocities[Spin.up][5, :, :, -3:])
        # print(velocities_no_sym[Spin.up][5, :, :, -3:])
        # print(sym_info["ir_to_full_idx"][-10:])

        np.testing.assert_array_almost_equal(
            energies[Spin.up], energies_no_sym[Spin.up], decimal=12
        )
        np.testing.assert_array_almost_equal(
            velocities[Spin.up], velocities_no_sym[Spin.up], decimal=12
        )
        np.testing.assert_array_almost_equal(
            curvature[Spin.up], curvature_no_sym[Spin.up], decimal=12
        )

        for l in projections[Spin.up]:
            np.testing.assert_array_almost_equal(
                projections[Spin.up][l], projections_no_sym[Spin.up][l]
            )

    def test_get_energies_interpolater(self):

        initialize_amset_logger()

        amset_data = self.interpolater.get_amset_data()

        energies, velocities, projections, sym_info = self.interpolater.get_energies(
            amset_data.kpoints,
            None,
            return_velocity=True,
            atomic_units=True,
            curvature=False,
            return_projections=True,
            symprec=0.1,
            return_vel_outer_prod=True,
            return_kpoint_mapping=True,
        )

        np.testing.assert_array_almost_equal(
            energies[Spin.up], amset_data.energies[Spin.up]
        )
        np.testing.assert_array_almost_equal(
            velocities[Spin.up], amset_data.velocities_product[Spin.up]
        )

        for l in projections[Spin.up]:
            np.testing.assert_array_almost_equal(
                projections[Spin.up][l], amset_data._projections[Spin.up][l]
            )

        np.testing.assert_array_equal(
            sym_info["ir_kpoints_idx"], amset_data.ir_kpoints_idx
        )
        np.testing.assert_array_equal(
            sym_info["ir_to_full_idx"], amset_data.ir_to_full_kpoint_mapping
        )



import logging
from abc import ABC, abstractmethod
from typing import Any, Dict, Tuple

import numpy as np
from pymatgen import Spin

from amset.constants import boltzmann_au, hbar, s_to_au
from amset.core.data import AmsetData
from amset.log import log_list
from amset.scattering.common import calculate_inverse_screening_length_sq

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

logger = logging.getLogger(__name__)


class AbstractInelasticScattering(ABC):

    name: str
    required_properties: Tuple[str]

    def __init__(self, properties, doping, temperatures, nbands):
        self.properties = properties
        self.doping = doping
        self.temperatures = temperatures
        self.nbands = nbands
        self.spins = list(nbands.keys())

    @classmethod
    def from_amset_data(
        cls, materials_properties: Dict[str, Any], amset_data: AmsetData
    ):
        return cls(
            cls.get_properties(materials_properties),
            amset_data.doping,
            amset_data.temperatures,
            cls.get_nbands(amset_data),
        )

    @abstractmethod
    def prefactor(self, spin: Spin, b_idx: int):
        pass

    @abstractmethod
    def factor(self, unit_q, norm_q_sq, emission, f):
        pass

    @classmethod
    def get_properties(cls, materials_properties):
        return {p: materials_properties[p] for p in cls.required_properties}

    @staticmethod
    def get_nbands(amset_data):
        return {s: len(amset_data.energies[s]) for s in amset_data.spins}


class PolarOpticalScattering(AbstractInelasticScattering):

    name = "POP"
    required_properties = (
        "pop_frequency",
        "static_dielectric",
        "high_frequency_dielectric",
        "free_carrier_screening",
    )

    def __init__(
        self,
        properties,
        doping,
        temperatures,
        nbands,
        pop_frequency,
        n_po,
        inverse_screening_length_sq,
    ):
        super().__init__(properties, doping, temperatures, nbands)
        self.pop_frequency = pop_frequency
        self.n_po = n_po
        self.inverse_screening_length_sq = inverse_screening_length_sq
        self._prefactor = s_to_au * pop_frequency / 2

    @classmethod
    def from_amset_data(
        cls, materials_properties: Dict[str, Any], amset_data: AmsetData
    ):
        logger.info("Initializing POP scattering")

        # convert from THz to angular frequency in Hz
        pop_frequency = (
            materials_properties["pop_frequency"] * 1e12 * 2 * np.pi / s_to_au
        )

        if materials_properties["free_carrier_screening"]:
            # use high-frequency diel for screening length
            avg_diel = np.linalg.eigvalsh(
                materials_properties["high_frequency_dielectric"]
            ).mean()
            inverse_screening_length_sq = calculate_inverse_screening_length_sq(
                amset_data, avg_diel
            )
        else:
            inverse_screening_length_sq = np.zeros_like(amset_data.fermi_levels)

        # n_po (phonon concentration) has shape (ntemps, )
        n_po = 1 / (
            np.exp(pop_frequency / (boltzmann_au * amset_data.temperatures)) - 1
        )

        n_po = n_po[None, :]

        log_list(
            [
                "average N_po: {:.4f}".format(np.mean(n_po)),
                "ω_po: {:.4g} 2π THz".format(
                    materials_properties["pop_frequency"] * 2 * np.pi
                ),
                "ħω: {:.4f} eV".format(pop_frequency * hbar * s_to_au),
            ]
        )

        # want to store two intermediate properties for:
        #             emission      and        absorption
        # (1-f)(N_po + 1) + f(N_po) and (1-f)N_po + f(N_po + 1)
        # note that these are defined for the scattering rate S(k', k).
        # For the rate S(k, k') the definitions are reversed.

        # self.emission_f_out = {
        #     s: n_po + 1 - amset_data.f[s]
        #     for s in amset_data.spins}
        # self.absorption_f_out = {
        #     s: n_po + amset_data.f[s]
        #     for s in amset_data.spins}
        #
        # self.emission_f_in = {
        #     s: n_po + amset_data.f[s]
        #     for s in amset_data.spins}
        # self.absorption_f_in = {
        #     s: n_po + 1 - amset_data.f[s]
        #     for s in amset_data.spins}
        return cls(
            cls.get_properties(materials_properties),
            amset_data.doping,
            amset_data.temperatures,
            cls.get_nbands(amset_data),
            pop_frequency,
            n_po,
            inverse_screening_length_sq,
        )

    def prefactor(self, spin: Spin, b_idx: int):
        # need to return prefactor with shape (nspins, ndops, ntemps, nbands)
        return self._prefactor * np.ones((len(self.doping), len(self.temperatures)))

    def factor(self, unit_q: np.ndarray, norm_q_sq: np.ndarray, emission, f):
        # presuming that this is out scattering
        if emission:
            factor = self.n_po + 1 - f
        else:
            factor = self.n_po + f

        high_t = self.properties["high_frequency_dielectric"]
        static_t = self.properties["static_dielectric"]
        high_freq_diel = np.einsum("ij,ij->i", unit_q, np.dot(high_t, unit_q.T).T)
        static_diel = np.einsum("ij,ij->i", unit_q, np.dot(static_t, unit_q.T).T)

        dielectric_term = 4 * np.pi * (1 / high_freq_diel - 1 / static_diel)

        return (
            factor[..., None]
            * dielectric_term[None, None]
            / (norm_q_sq[None, None] + self.inverse_screening_length_sq[..., None])
        )

        # # factor should have shape (ndops, ntemps, nkpts)
        # factor = 1 / np.tile(k_diff_sq, (len(self.doping),
        #                                  len(self.temperatures), 1))
        # if emission:
        #     if out:
        #         return factor * self.emission_f_out[spin][:, :, b_idx, k_idx]
        #     else:
        #         return factor * self.emission_f_in[spin][:, :, b_idx, k_idx]
        # else:
        #     if out:
        #         return factor * self.absorption_f_out[spin][:, :, b_idx, k_idx]
        #     else:
        #         return factor * self.absorption_f_in[spin][:, :, b_idx, k_idx]



import click
from click import argument, option

from amset.tools.common import path_type, zero_weighted_type
from amset.util import parse_doping, parse_temperatures

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"

temp_doping_defaults = {"temperatures": [300], "doping": [-1e15, 1e15]}


@click.command()
@argument("filename", type=path_type)
@option(
    "-i",
    "--interpolation-factor",
    type=float,
    help="band structure interpolation factor",
)
@option(
    "-d",
    "--doping",
    metavar="D",
    type=parse_doping,
    help='doping concentrations (e.g. "1E13,1E14" or "1E13:1E20:8")',
)
@option(
    "-t",
    "--temperatures",
    metavar="T",
    type=parse_temperatures,
    help='temperatures (e.g. "300,400" or "300:1000:8")',
)
@option("--scissor", type=float, help="amount to scissor the band gap")
@option("--bandgap", type=float, help="manually set the band gap")
@option("--average/--no-average", default=False, help="report the harmonic average")
@option("--print-log/--no-log", default=True, help="whether to print log messages")
@option(
    "-z",
    "--zero-weighted-kpoints",
    help="how to process zero-weighted k-points",
    type=zero_weighted_type,
)
def eff_mass(filename, **kwargs):
    """
    Calculate conductivity effective mass.
    """
    import numpy as np
    from scipy import constants
    from tabulate import tabulate

    from amset.constants import bohr_to_cm, defaults
    from amset.core.run import Runner

    settings = {
        "scattering_type": ["CRT"],
        "constant_relaxation_time": 1e-14,
        "use_projections": True,
        "calculate_mobility": False,
        "separate_mobility": False,
        "write_log": False,
    }
    for setting in defaults:
        if setting in kwargs and kwargs[setting] is not None:
            settings[setting] = kwargs[setting]
        elif setting == "doping" or setting == "temperatures":
            settings[setting] = temp_doping_defaults[setting]

    runner = Runner.from_vasprun(filename, settings)
    amset_data = runner.run()
    inv_cond = np.linalg.inv(amset_data.conductivity)
    doping_scale = 1 / bohr_to_cm ** 3
    doping = (np.abs(amset_data.doping) * doping_scale)[:, None, None, None]
    crt = settings["constant_relaxation_time"]

    # use eigvals rather than eigvalsh to avoid sorting the eigenvalues
    masses = inv_cond * crt * doping * 10 ** 6 * constants.e ** 2 / constants.m_e
    eigs = np.linalg.eigvals(masses).real

    mass_info = []
    for n, t in np.ndindex(amset_data.fermi_levels.shape):
        info = [amset_data.doping[n] * doping_scale, amset_data.temperatures[t]]
        if kwargs["average"]:
            info.append((3 / (1 / eigs[n, t]).sum()))
        else:
            info.extend(eigs[n, t].tolist())
        mass_info.append(info)

    headers = ["conc [cm⁻³]", "temp [K]"]
    floatfmt = [".2e", ".1f"]

    if kwargs["average"]:
        headers.append("m*")
        floatfmt.append(".3f")
    else:
        headers.extend(["mₓ*", "mᵧ*", "m₂*"])
        floatfmt.extend([".3f"] * 3)

    table = tabulate(
        mass_info,
        headers=headers,
        numalign="right",
        stralign="center",
        floatfmt=floatfmt,
    )

    click.echo("\nEffective masses:\n")
    click.echo(table)

    return mass_info



# import warnings
# from copy import deepcopy
# from pathlib import Path
#
# import click
# import numpy as np
#
# __author__ = "Alex Ganose"
# __maintainer__ = "Alex Ganose"
# __email__ = "aganose@lbl.gov"
#
# from pymatgen import Spin
#
# warnings.filterwarnings("ignore", category=UserWarning, module="pymatgen")
#
#
# @click.command(context_settings=dict(help_option_names=["-h", "--help"]))
# @click.argument("sym-folder")
# @click.argument("no-sym-folder")
# @click.option("--symprec", type=float, default=1e-2, help="symmetry precision")
# @click.option("--basic", is_flag=True, default=False, help="use basic mode")
# def desym(sym_folder, no_sym_folder, symprec, basic):
#     """Test desymmetrization routines"""
#     from pymatgen.io.vasp import Vasprun, Wavecar
#
#     from amset.electronic_structure.symmetry import expand_kpoints
#     from amset.log import initialize_amset_logger
#     from amset.wavefunction.common import desymmetrize_coefficients
#     from amset.wavefunction.io import load_coefficients
#
#     initialize_amset_logger(filename=False, print_log=True)
#
#     sym_folder = Path(sym_folder)
#     no_sym_folder = Path(no_sym_folder)
#
#     click.echo("Loading symmetrized wavefunction")
#
#     coeffs_sym, gpoints, kpoints, structure = load_coefficients(sym_folder)
#     full_kpoints, *symmetry_mapping = expand_kpoints(
#         structure, kpoints, time_reversal=True, return_mapping=True, symprec=symprec
#     )
#
#     coeffs_desym = desymmetrize_coefficients(
#         coeffs_sym, gpoints, kpoints, *symmetry_mapping
#     )
#
#     click.echo("Loading unsymmetrized wavefunction")
#     coeffs_no_sym, gpoints_no_sym, no_sym_kpoints, _ = load_coefficients(no_sym_folder)
#     if set(map(tuple, gpoints)) != set(map(tuple, gpoints_no_sym)):
#         raise ValueError("Gpoints do not match")
#
#     # map desym k-points to same order as sym k-points
#     desym_sort_idx = get_sort_idx(full_kpoints)
#     no_sym_sort_idx = get_sort_idx(no_sym_kpoints)
#
#     coeffs_sym_orig = deepcopy(coeffs_sym)
#
#     coeffs_desym_order = {s: c[:, desym_sort_idx] for s, c in coeffs_desym.items()}
#     coeffs_no_sym_order = {s: c[:, no_sym_sort_idx] for s, c in coeffs_no_sym.items()}
#
#     data = check_overlap(coeffs_desym_order, coeffs_no_sym_order)
#     # analyze_bad_overlaps(kpoints, no_sym_kpoints, gpoints, gpoints_no_sym, coeffs_sym_orig, coeffs_no_sym, data, symmetry_mapping, desym_sort_idx)
#     analyze_bad_overlaps(
#         kpoints,
#         full_kpoints,
#         no_sym_kpoints,
#         gpoints,
#         gpoints_no_sym,
#         coeffs_desym,
#         coeffs_no_sym,
#         data,
#         symmetry_mapping,
#         desym_sort_idx,
#     )
#
#
# # @click.command(context_settings=dict(help_option_names=["-h", "--help"]))
# # @click.argument("sym-folder")
# # @click.argument("no-sym-folder")
# # @click.option("--symprec", type=float, default=1e-2, help="symmetry precision")
# # @click.option("--basic", is_flag=True, default=False, help="use basic mode")
# # def desym(sym_folder, no_sym_folder, symprec, basic):
# #     """Test desymmetrization routines"""
# #     from pymatgen.io.vasp import Vasprun, Wavecar
# #
# #     from amset.electronic_structure.symmetry import expand_kpoints
# #     from amset.log import initialize_amset_logger
# #     from amset.wavefunction.common import desymmetrize_coefficients
# #     from amset.wavefunction.vasp import get_wavefunction_coefficients
# #
# #     initialize_amset_logger(filename=False, print_log=True)
# #
# #     sym_folder = Path(sym_folder)
# #     no_sym_folder = Path(no_sym_folder)
# #
# #     click.echo("Loading symmetrized wavefunction")
# #     wave_sym = Wavecar(sym_folder / "WAVECAR")
# #
# #     vr = Vasprun(sym_folder / "vasprun.xml")
# #     structure = vr.final_structure
# #     kpoints = np.array(wave_sym.kpoints)
# #
# #     if basic:
# #         click.echo("Loading unsymmetrized wavefunction")
# #         wave_no_sym = Wavecar(no_sym_folder / "WAVECAR")
# #         desym_basic(structure, wave_sym, wave_no_sym)
# #         import sys
# #
# #         sys.exit()
# #
# #     full_kpoints, *symmetry_mapping = expand_kpoints(
# #         structure, kpoints, time_reversal=True, return_mapping=True, symprec=symprec
# #     )
# #
# #     coeffs_sym, gpoints = get_wavefunction_coefficients(wave_sym, encut=wave_sym.encut)
# #
# #     coeffs_desym = desymmetrize_coefficients(
# #         coeffs_sym, gpoints, kpoints, *symmetry_mapping
# #     )
# #
# #     click.echo("Loading unsymmetrized wavefunction")
# #     # wave_no_sym = Wavecar(no_sym_folder / "WAVECAR")
# #     #
# #     # coeffs_no_sym, gpoints_no_sym = get_wavefunction_coefficients(
# #     #     wave_no_sym, encut=wave_sym.encut
# #     # )
# #     from amset.wavefunction.io import load_coefficients
# #     coeffs_no_sym, gpoints_no_sym, no_sym_kpoints, _ = load_coefficients(no_sym_folder)
# #     if set(map(tuple, gpoints)) != set(map(tuple, gpoints_no_sym)):
# #         raise ValueError("Gpoints do not match")
# #
# #     # map desym k-points to same order as sym k-points
# #     desym_sort_idx = get_sort_idx(full_kpoints)
# #     no_sym_sort_idx = get_sort_idx(no_sym_kpoints)
# #
# #     coeffs_desym = {s: c[:, desym_sort_idx] for s, c in coeffs_desym.items()}
# #     coeffs_no_sym = {s: c[:, no_sym_sort_idx] for s, c in coeffs_no_sym.items()}
# #
# #     data = check_overlap(coeffs_desym, coeffs_no_sym)
# #     analyze_bad_overlaps(wave_sym, None, data, symmetry_mapping, desym_sort_idx)
#
#
# def analyze_bad_overlaps(
#     orig_kpoints,
#     sym_kpoints,
#     nosym_kpoints,
#     sym_gpoints,
#     nosym_gpoints,
#     sym_coeffs,
#     nosym_coeffs,
#     data,
#     symmetry_mapping,
#     sort_idx,
# ):
#     rotations, translations, is_tr, op_mapping, kp_mapping = symmetry_mapping
#     kp_mapping = kp_mapping[sort_idx]
#     op_mapping = op_mapping[sort_idx]
#
#     rots = rotations[op_mapping]
#     tau = translations[op_mapping]
#
#     for spin, spin_data in data.items():
#         if len(spin_data["kpoints"]) == 0:
#             continue
#         bad_kpoints = np.array(spin_data["kpoints"])
#         bad_bands = np.array(spin_data["bands"])
#         diffs = np.array(spin_data["diffs"])
#         # original_k_idxs = kp_mapping[bad_kpoints]
#         # original_k_idxs, unique_idx = np.unique(original_k_idxs, return_index=True)
#         # bad_bands = bad_bands[unique_idx]
#
#         # print(bad_kpoints)
#         # print(len(original_k_idxs), len(bad_kpoints))
#
#         count_actually_bad = 0
#         # for nk, nb, nop, diff in zip(
#         #         original_k_idxs, bad_bands, bad_kpoints[unique_idx], diffs[unique_idx]
#         # ):
#         #     kpoint = orig_kpoints[nk]
#         for nk, nb, nop, diff in zip(bad_kpoints, bad_bands, bad_kpoints, diffs):
#             kpoint = sym_kpoints[sort_idx][nk]
#             original_k = orig_kpoints[kp_mapping][nk]
#             overlap = test_original_overlap(
#                 sym_kpoints,
#                 nosym_kpoints,
#                 sym_gpoints,
#                 nosym_gpoints,
#                 sym_coeffs,
#                 nosym_coeffs,
#                 kpoint,
#                 nb,
#             )
#
#             final_k = np.dot(rots[nop], original_k)
#             kdiff = np.around(final_k)
#             final_k -= kdiff
#
#             edges = np.around(final_k, 5) == -0.5
#             final_k += edges
#             print(original_k.round(4), "->", final_k.round(4))
#             print("  kpoints", kpoint)
#             print("  kdiff", kdiff)
#             print("  edges", edges)
#             print("  rot was", rots[nop].ravel())
#             print("  tau was", tau[nop].ravel())
#             print(" ", overlap)
#
#         print(
#             "{} out of {} are bad due to VASP".format(
#                 count_actually_bad, len(bad_bands)
#             )
#         )
#
#
# # def analyze_bad_overlaps(wave_sym, wave_no_sym, data, symmetry_mapping, sort_idx):
# #     rotations, translations, is_tr, op_mapping, kp_mapping = symmetry_mapping
# #     kp_mapping = kp_mapping[sort_idx]
# #     op_mapping = op_mapping[sort_idx]
# #
# #     rots = rotations[op_mapping]
# #     tau = translations[op_mapping]
# #
# #     for spin, spin_data in data.items():
# #         if len(spin_data["kpoints"]) == 0:
# #             continue
# #         bad_kpoints = np.array(spin_data["kpoints"])
# #         bad_bands = np.array(spin_data["bands"])
# #         diffs = np.array(spin_data["diffs"])
# #         original_k_idxs = kp_mapping[bad_kpoints]
# #         original_k_idxs, unique_idx = np.unique(original_k_idxs, return_index=True)
# #         bad_bands = bad_bands[unique_idx]
# #
# #         print(bad_kpoints)
# #         print(len(original_k_idxs), len(bad_kpoints))
# #
# #         count_actually_bad = 0
# #         for nk, nb, nop, diff in zip(
# #             original_k_idxs, bad_bands, bad_kpoints[unique_idx], diffs[unique_idx]
# #         ):
# #             kpoint = wave_sym.kpoints[nk]
# #             overlap = test_original_overlap(wave_sym, wave_no_sym, kpoint, nb)
# #             final_k = np.dot(rots[nop], kpoint)
# #             final_k -= np.round(final_k)
# #             print(kpoint.round(4), "->", final_k)
# #             if abs(overlap - 1) > 0.005:
# #                 print("  bad in raw data")
# #                 count_actually_bad += 1
# #             else:
# #                 print("  ok in raw data")
# #                 print("  rot was", rots[nop].ravel())
# #                 print("  tau was", tau[nop].ravel())
# #             print(" ", overlap, diff)
# #
# #         print(
# #             "{} out of {} are bad due to VASP".format(
# #                 count_actually_bad, len(bad_bands)
# #             )
# #         )
#
#
# def check_overlap(coeffs_a, coeffs_b):
#     from tqdm.auto import tqdm
#
#     band_idx = 1
#
#     data = {s: {"kpoints": [], "diffs": [], "bands": []} for s in coeffs_a.keys()}
#     for spin, spin_coeffs_a in coeffs_a.items():
#         spin_coeffs_b = coeffs_b[spin]
#
#         if spin_coeffs_a.shape != spin_coeffs_b.shape:
#             raise ValueError("Coefficients shapes do not match")
#
#         diffs = []
#         for nb, nk in tqdm(list(np.ndindex(spin_coeffs_a.shape[:2]))):
#             if nb != band_idx:
#                 continue
#             v_a = spin_coeffs_a[nb, nk]
#             v_b = spin_coeffs_b[nb, nk]
#
#             diff = abs(np.vdot(v_a, v_b))
#             diffs.append(diff)
#
#             if abs(diff - 1) > 0.005:
#                 data[spin]["kpoints"].append(nk)
#                 data[spin]["bands"].append(nb)
#                 data[spin]["diffs"].append(diff)
#
#         diffs = np.array(diffs)
#         bad_overlap = np.abs(diffs - 1) > 0.1
#         n_bad = np.sum(bad_overlap)
#         click.echo("{:.1f} k-points are bad in band {}".format(n_bad, band_idx))
#     return data
#
#
# def get_sort_idx(kpoints):
#     kpoints = np.array(kpoints).round(5)
#     return np.lexsort((kpoints[:, 2], kpoints[:, 1], kpoints[:, 0]))
#
#
# def desym_basic(structure, wave_sym, wave_no_sym):
#     from amset.electronic_structure.symmetry import expand_kpoints
#
#     (
#         full_kpoints,
#         rotations,
#         translations,
#         is_tri,
#         op_mapping,
#         kp_mapping,
#     ) = expand_kpoints(
#         structure, wave_sym.kpoints, time_reversal=True, return_mapping=True
#     )
#
#     def get_sort_idx_basic(g):
#         g = np.array(g).round(5)
#         g[g == -0.5] = 0.5
#         return np.lexsort((g[:, 2], g[:, 1], g[:, 0]))
#
#     nbmax = wave_sym._nbmax
#     encut = wave_sym.encut
#     wave_const = 0.262465831
#     band_idx = 0
#
#     ops = rotations[op_mapping]
#     taus = translations[op_mapping]
#     tris = is_tri[op_mapping]
#     diffs = []
#     for k_idx in list(range(len(ops))):
#
#         map_idx = kp_mapping[k_idx]
#         op = ops[k_idx]
#         tau = taus[k_idx]
#         kpoint = wave_sym.kpoints[map_idx]
#         gpoints = wave_sym.Gpoints[map_idx]
#         coeffs = wave_sym.coeffs[map_idx]
#
#         rot_kpoint = np.dot(op, kpoint)
#
#         kdiff = np.around(rot_kpoint)
#         rot_kpoint -= kdiff
#
#         edges = np.around(rot_kpoint, 5) == -0.5
#         rot_kpoint += edges
#         kdiff -= edges
#
#         all_g = np.array(list(np.ndindex(tuple(2 * nbmax[::-1] + 1))))
#         all_g = all_g[:, [2, 1, 0]]  # swap columns
#         all_g[all_g[:, 2] > nbmax[2], 2] -= 2 * nbmax[2] + 1
#         all_g[all_g[:, 1] > nbmax[1], 1] -= 2 * nbmax[1] + 1
#         all_g[all_g[:, 0] > nbmax[0], 0] -= 2 * nbmax[0] + 1
#
#         cart_g = np.dot(all_g + rot_kpoint, structure.lattice.reciprocal_lattice.matrix)
#         norm_g = np.linalg.norm(cart_g, axis=1)
#         ener_g = norm_g ** 2 / wave_const
#
#         valid_g = all_g[ener_g <= encut]
#
#         min_g = valid_g.min(axis=0)
#         max_g = valid_g.max(axis=0)
#         ng = max_g - min_g + 1
#
#         shifted_valid_g = valid_g - min_g
#         valid_indices = (
#             shifted_valid_g[:, 0] * ng[1] * ng[2]
#             + shifted_valid_g[:, 1] * ng[2]
#             + shifted_valid_g[:, 2]
#         )
#
#         rot_gpoints = np.dot(op, gpoints.T).T
#         rot_gpoints = np.around(rot_gpoints).astype(int)
#         rot_gpoints += kdiff.astype(int)
#
#         shifted_rot_g = rot_gpoints - min_g
#         rot_indices = (
#             shifted_rot_g[:, 0] * ng[1] * ng[2]
#             + shifted_rot_g[:, 1] * ng[2]
#             + shifted_rot_g[:, 2]
#         )
#
#         if set(rot_indices) != set(valid_indices):
#             print("BAD PLANE WAVE MAPPING")
#
#         f = np.dot(rot_kpoint, tau) + np.dot(rot_gpoints, tau)
#         exp_factor = np.exp(-1j * 2 * np.pi * f)
#         new_coeffs = exp_factor[None, :] * coeffs
#         if tris[k_idx]:
#             new_coeffs = np.conjugate(new_coeffs)
#
#         nosym_kp = get_sort_idx_basic(np.array(wave_no_sym.kpoints))[k_idx]
#
#         nosym_gpoints = wave_no_sym.Gpoints[nosym_kp].astype(int)
#         shifted_nosym_g = nosym_gpoints - min_g
#         nosym_indices = (
#             shifted_nosym_g[:, 0] * ng[1] * ng[2]
#             + shifted_nosym_g[:, 1] * ng[2]
#             + shifted_nosym_g[:, 2]
#         )
#
#         if set(rot_indices) != set(nosym_indices):
#             print("BAD PLANE WAVE MAPPING2")
#
#         v1 = new_coeffs[band_idx][np.argsort(rot_indices)]
#         v2 = wave_no_sym.coeffs[nosym_kp][band_idx][np.argsort(nosym_indices)]
#
#         o1 = abs(np.vdot(v1, v2))
#         o2 = abs(np.vdot(v2, v2))
#         diffs.append(o1 / o2)
#         if abs(o1 / o2 - 1) > 0.05:
#             print(kdiff, edges, np.dot(op, kpoint), taus[k_idx])
#             print(ops[k_idx])
#
#     diffs = np.array(diffs)
#     bad_overlap = np.abs(diffs - 1) > 0.1
#     n_bad = np.sum(bad_overlap)
#     total = n_bad
#     click.echo("{:.1f} k-points are bad".format(total))
#
#
# def test_original_overlap(
#     sym_kpoints,
#     nosym_kpoints,
#     sym_gpoints,
#     nosym_gpoints,
#     sym_coeffs,
#     nosym_coeffs,
#     kpoint,
#     band_idx,
# ):
#     sym_diff = np.linalg.norm(np.array(sym_kpoints) - kpoint, axis=1)
#     if np.min(sym_diff) > 0.01:
#         raise ValueError("k-point is not in symmetrized wavefunction")
#     sym_k_idx = sym_diff.argmin()
#
#     nosym_diff = np.linalg.norm(np.array(nosym_kpoints) - kpoint, axis=1)
#     if np.min(nosym_diff) > 0.01:
#         raise ValueError("k-point is not in no-symmetry wavefunction")
#     nosym_k_idx = nosym_diff.argmin()
#
#     gdiff = np.abs(sym_gpoints - nosym_gpoints)
#     if gdiff.max() > 0.2:
#         ValueError("Gpoints differ!")
#
#     v1 = sym_coeffs[Spin.up][band_idx, sym_k_idx]
#     v1 /= np.linalg.norm(v1)
#
#     v2 = nosym_coeffs[Spin.up][band_idx, nosym_k_idx]
#     v2 /= np.linalg.norm(v2)
#
#     return abs(np.vdot(v1, v2))
#
#
# # def test_original_overlap(wave_sym, wave_nosym, kpoint, band_idx):
# #     sym_diff = np.linalg.norm(np.array(wave_sym.kpoints) - kpoint, axis=1)
# #     if np.min(sym_diff) > 0.01:
# #         raise ValueError("k-point is not in symmetrized wavefunction")
# #     sym_k_idx = sym_diff.argmin()
# #
# #     nosym_diff = np.linalg.norm(np.array(wave_nosym.kpoints) - kpoint, axis=1)
# #     if np.min(nosym_diff) > 0.01:
# #         raise ValueError("k-point is not in no-symmetry wavefunction")
# #     nosym_k_idx = nosym_diff.argmin()
# #
# #     gdiff = np.abs(wave_sym.Gpoints[sym_k_idx] - wave_nosym.Gpoints[nosym_k_idx])
# #     if gdiff.max() > 0.2:
# #         ValueError("Gpoints differ!")
# #
# #     v1 = wave_sym.coeffs[sym_k_idx][band_idx]
# #     v1 /= np.linalg.norm(v1)
# #
# #     v2 = wave_nosym.coeffs[nosym_k_idx][band_idx]
# #     v2 /= np.linalg.norm(v2)
# #
# #     return abs(np.vdot(v1, v2))



from setuptools import find_packages, setup

from amset import __version__

with open("README.md", "r") as file:
    long_description = file.read()

if __name__ == "__main__":
    setup(
        name="amset",
        version=__version__,
        description="AMSET is a tool to calculate carrier transport properties "
        "from ab initio calculation data",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/hackingmaterials/amset",
        author="Alex Ganose",
        author_email="aganose@lbl.gov",
        license="modified BSD",
        keywords="mobility conductivity seebeck scattering lifetime rates dft vasp",
        packages=find_packages(),
        package_data={
            "amset": [
                "defaults.yaml",
                "plot/amset_base.mplstyle",
                "plot/revtex.mplstyle",
            ]
        },
        data_files=["LICENSE"],
        zip_safe=False,
        install_requires=[
            "quadpy==0.16.2",
            "numpy==1.19.4",
            "pymatgen==2020.11.11",
            "scipy==1.5.4",
            "monty==4.0.2",
            "matplotlib==3.3.3",
            "BoltzTraP2==20.7.1",
            "tqdm==4.51.0",
            "tabulate==0.8.7",
            "memory_profiler==0.58.0",
            "spglib==1.16.0",
            "click==7.1.2",
            "sumo==2.1.0",
            "h5py==2.10.0",
            "pyFFTW==0.12.0",
            "interpolation==2.1.6",
        ],
        extras_require={
            "docs": [
                "mkdocs==1.1.2",
                "mkdocs-material==6.1.4",
                "mkdocs-minify-plugin==0.3.0",
                "mkdocs-macros-plugin==0.4.20",
                "markdown-include==0.6.0",
                "markdown-katex==202009.1026",
            ],
            "tests": ["pytest==6.1.2", "pytest-cov==2.10.1"],
            "all-electron": ["pawpyseed==0.6.3"],
            "dev": [
                "coverage==5.3",
                "codacy-coverage==1.3.11",
                "pycodestyle==2.6.0",
                "mypy==0.790",
                "pydocstyle==5.1.1",
                "flake8==3.8.4",
                "pylint==2.6.0",
                "black==20.8b1",
            ],
        },
        classifiers=[
            "Programming Language :: Python :: 3.6",
            "Development Status :: 4 - Beta",
            "Intended Audience :: Science/Research",
            "Intended Audience :: System Administrators",
            "Intended Audience :: Information Technology",
            "Operating System :: OS Independent",
            "Topic :: Other/Nonlisted Topic",
            "Topic :: Scientific/Engineering",
        ],
        tests_require=["pytest"],
        entry_points={
            "console_scripts": [
                "amset = amset.tools.cli:cli",
                "desym = amset.tools.desym:desym",
            ]
        },
    )



import matplotlib.pyplot

from amset.plot.base import (  # noqa
    BaseMeshPlotter,
    BaseTransportPlotter,
    amset_base_style,
    revtex_style,
)

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"


def styled_plot(*style_sheets):
    """Return a decorator that will apply matplotlib style sheets to a plot.

    ``style_sheets`` is a base set of styles, which will be ignored if
    ``no_base_style`` is set in the decorated function arguments.

    The style will further be overwritten by any styles in the ``style``
    optional argument of the decorated function.

    Args:
        style_sheets (:obj:`list`, :obj:`str`, or :obj:`dict`): Any matplotlib
            supported definition of a style sheet. Can be a list of style of
            style sheets.
    """

    def decorator(get_plot):
        def wrapper(*args, fonts=None, style=None, no_base_style=False, **kwargs):

            if no_base_style:
                list_style = []
            else:
                list_style = list(style_sheets)

            if style is not None:
                if "revtex" in style:
                    list_style = [revtex_style]
                elif isinstance(style, list):
                    list_style += style
                else:
                    list_style += [style]

            if fonts is not None:
                list_style += [{"font.family": "sans-serif", "font.sans-serif": fonts}]

            matplotlib.pyplot.style.use(list_style)
            return get_plot(*args, **kwargs)

        return wrapper

    return decorator



"""
This module contains a script for using amset from the command line.
"""

from collections import defaultdict
from typing import Any, Dict

import click
from click import option

from amset.tools.common import zero_weighted_type
from amset.util import parse_deformation_potential, parse_doping, parse_temperatures

__author__ = "Alex Ganose"
__maintainer__ = "Alex Ganose"
__email__ = "aganose@lbl.gov"


@click.command()
@option(
    "--directory", default=".", help="path to directory with vasprun and settings files"
)
@option("-v", "--vasprun", help="path to vasprun.xml file")
@option("-s", "--settings", help="path to settings file")
@option(
    "-i",
    "--interpolation-factor",
    type=float,
    help="band structure interpolation factor",
)
@option(
    "--d",
    "--doping",
    metavar="D",
    type=parse_doping,
    help='doping concentrations (e.g. "1E13,1E14" or "1E13:1E20:8")',
)
@option(
    "--t",
    "--temperatures",
    metavar="T",
    type=parse_temperatures,
    help='temperatures (e.g. "300,400" or "300:1000:8")',
)
@option("--scissor", type=float, help="amount to scissor the band gap")
@option("--bandgap", type=float, help="manually set the band gap")
@option(
    "-z",
    "--zero-weighted-kpoints",
    help="how to process zero-weighted k-points",
    type=zero_weighted_type,
)
@option(
    "--free-carrier-screening/--no-free-carrier-screening",
    default=None,
    help="whether free carriers will screen POP and PIE rates [default: False]",
)
@option("--scattering-type", type=float, help="scattering mechanisms to include")
@option(
    "--high-frequency-dielectric", type=float, help="high-frequency dielectric constant"
)
@option("--static-dielectric", type=float, help="static dielectric constant")
@option("--elastic-constant", type=float, help="elastic constant [GPa]")
@option(
    "--deformation-potential",
    metavar="D",
    type=parse_deformation_potential,
    help='deformation potential [eV] (e.g. "7.4" or "7.4,6.8")',
)
@option("--piezoelectric-constant", type=float, help="piezoelectric constant")
@option("--acceptor-charge", type=float, help="acceptor defect charge")
@option("--donor-charge", type=float, help="donor defect charge")
@option("--pop-frequency", type=float, help="polar optical phonon frequency [THz]")
@option("--mean-free-path", type=float, help="set the mean free path of electrons [nm]")
@option("--constant-relaxation-time", type=float, help="constant relaxation time [s]")
@option(
    "--energy-cutoff",
    type=float,
    help="energy cut-off for band electronic_structure [eV]",
)
@option(
    "--fd-tol",
    type=float,
    help="Fermi-Dirac tolerance below which scattering rates are not "
    "calculated [%%]",
)
@option(
    "--cache-wavefunction/--no-cache-wavefunction",
    default=None,
    help="cache wavefunction coefficients; beware increased memory usage [default: True]",
)
@option("--dos-estep", type=float, help="dos energy step [eV]")
@option("--symprec", type=float, help="symmetry precision")
@option("--nworkers", type=int, help="number of processors to use")
@option(
    "--calculate-mobility/--no-calculate-mobility",
    default=None,
    help="whether to calculate mobility",
)
@option(
    "--separate-mobility/--no-separate-mobility",
    default=None,
    help="whether to separate the individual scattering rate mobilities",
)
@option("--file-format", help="output file format [options: json, yaml, txt, dat]")
@option(
    "--write-input/--no-write-input", default=None, help="write input settings to file"
)
@option(
    "--write-mesh/--no-write-mesh",
    default=None,
    help="write mesh data, including band energies and scattering rates",
)
@option("--print-log/--no-log", default=True, help="whether to print log messages")
def run(**kwargs):
    """
    Run AMSET on an ab initio band structure
    """

    from amset.constants import defaults
    from amset.core.run import Runner
    from amset.log import initialize_amset_logger

    if kwargs["print_log"] is not False:
        initialize_amset_logger()

    settings_override: Dict[str, Dict[str, Any]] = defaultdict(dict)
    for setting in defaults:
        if setting in kwargs and kwargs[setting] is not None:
            settings_override[setting] = kwargs[setting]

    runner = Runner.from_directory(
        directory=kwargs["directory"],
        vasprun=kwargs["vasprun"],
        settings_file=kwargs["settings"],
        settings_override=settings_override,
    )

    runner.run()
